# Copyright 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import binascii
import ctypes
from ctypes import (
    CFUNCTYPE, Structure, c_uint8, c_uint32, c_long, c_void_p,
    c_uint64, c_char_p, c_int, c_uint16, POINTER, Union, byref)
import json
import random
import signal
import subprocess
import sys
import time
import yaml
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import (
    WjhDropRawInfo,
    WjhL1DropAggregate,
    WjhDropAggregate)
from scapy.all import (raw, Ether, IP, IPv6, TCP, UDP)
from netq_lib.common.utils import subprocess_wrapper

AGENTSTATUS_SUCCESS = 0
AGENTSTATUS_FAILURE = 1

WJH_USER_CHANNEL_TAILDROP_E = 0
WJH_USER_CHANNEL_CYCLIC_E = 1
WJH_USER_CHANNEL_AGGREGATE_E = 2

WJH_DROP_REASON_GROUP_BUFFER_E = 0
WJH_DROP_REASON_GROUP_ACL_E = 1
WJH_DROP_REASON_GROUP_L1_E = 2
WJH_DROP_REASON_GROUP_L2_E = 3
WJH_DROP_REASON_GROUP_ROUTER_E = 4
WJH_DROP_REASON_GROUP_TUNNEL_E = 5

WJH_USER_CHANNEL_MODE_PUSH_E = 0
WJH_USER_CHANNEL_MODE_PULL_E = 1

WJH_SEVERITY_ALL_E = 0
WJH_SEVERITY_NOTICE_E = 1
WJH_SEVERITY_WARNING_E = 2
WJH_SEVERITY_ERROR_E = 3

severity_to_str = {
    WJH_SEVERITY_NOTICE_E: 'Notice',
    WJH_SEVERITY_WARNING_E: 'Warning',
    WJH_SEVERITY_ERROR_E: 'Error'
}

WJH_ACL_BINDING_POINT_INGRESS = 0
WJH_ACL_BINDING_POINT_EGRESS = 1
WJH_ACL_BINDING_POINT_RIF_INGRESS = 2
WJH_ACL_BINDING_POINT_RIF_EGRESS = 3

acl_bp_to_str = {
    WJH_ACL_BINDING_POINT_INGRESS: 'Ingress ACL',
    WJH_ACL_BINDING_POINT_EGRESS: 'Egress ACL',
    WJH_ACL_BINDING_POINT_RIF_INGRESS: 'RIF Ingress ACL',
    WJH_ACL_BINDING_POINT_RIF_EGRESS: 'RIF Egress ACL'
}

WJH_DROP_REASON_ID_NON_ROUTABLE_PACKET_E = 301
WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX0_16_E = 305
WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX1_16_E = 306
WJH_DROP_REASON_ID_NON_IP_PACKET_E = 307

WJH_DROP_REASON_ID_MLAG_PORT_ISOLATION_E = 201
WJH_DROP_REASON_ID_INGRESS_SPANNING_TREE_FILTER_E = 205

default_filters = [WJH_DROP_REASON_ID_NON_ROUTABLE_PACKET_E,
                   WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX0_16_E,
                   WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX1_16_E,
                   WJH_DROP_REASON_ID_NON_IP_PACKET_E,
                   WJH_DROP_REASON_ID_MLAG_PORT_ISOLATION_E,
                   WJH_DROP_REASON_ID_INGRESS_SPANNING_TREE_FILTER_E]

WjhDeactiveCb = CFUNCTYPE(None)


class WjhInitParam(Structure):
    _fields_ = [("force", c_uint8), ("deactive_cb", WjhDeactiveCb), ("bandwidth_percent", c_uint8),
                ("log_cb", c_void_p), ("conf_xml_path", c_char_p)]


class WjhDropCounter(Structure):
    _fields_ = [("dropped", c_uint64), ("received", c_uint64)]


class WjhTimespec(Structure):
    _fields_ = [("tv_sec", c_long), ("tv_nsec", c_long)]


class WjhAggregateTimestamp(Structure):
    _fields_ = [("first_timestamp", WjhTimespec), ("last_timestamp", WjhTimespec)]


class WjhDropReason(Structure):
    _fields_ = [("id", c_uint32),
                ("reason", c_char_p),
                ("severity", c_int),
                ("description", c_char_p)]


class WjhDropBasicAttr(Structure):
    _fields_ = [("recirculation_port", c_uint32),
                ("truncate_size", c_uint16)]


class WjhDropAdvancedAttr(Structure):
    _fields_ = [("trap_probability", c_uint32),
                ("trap_id", c_uint32),
                ("ptp_disabled", c_uint32)]


class WjhBufferDropAttr(Structure):
    _fields_ = [("basic_attr", WjhDropBasicAttr),
                ("advanced_attr", WjhDropAdvancedAttr)]


class WjhDropReasonGroupAttrUnion(Union):
    _fields_ = [("buffer_drop", WjhBufferDropAttr)]


class WjhDropReasonGroupAttr(Structure):
    _fields_ = [("attr", WjhDropReasonGroupAttrUnion)]


class WjhL2DropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhL2DropRawCb = CFUNCTYPE(c_int, POINTER(WjhL2DropRawInfo), POINTER(c_uint32))


class WjhRouterDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhRouterDropRawCb = CFUNCTYPE(c_int, POINTER(WjhRouterDropRawInfo), POINTER(c_uint32))


class WjhTunnelDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhTunnelDropRawCb = CFUNCTYPE(c_int, POINTER(WjhTunnelDropRawInfo), POINTER(c_uint32))


class WjhBufferDropRawInfo(Structure):
    _fields_ = [("packet", ctypes.c_void_p),
                ("packet_size", ctypes.c_uint32),
                ("ingress_port", ctypes.c_uint32),
                ("egress_port", ctypes.c_uint32),
                ("tc", ctypes.c_uint8),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", ctypes.c_uint8),
                ("ingress_lag", ctypes.c_uint32),
                ("egress_port_valid", ctypes.c_uint8),
                ("is_egress_lag_member", ctypes.c_uint8),
                ("egress_lag", ctypes.c_uint32),
                ("original_occupancy", ctypes.c_uint16),
                ("original_latency", ctypes.c_uint32)]


WjhBufferDropRawCb = CFUNCTYPE(c_int, POINTER(WjhBufferDropRawInfo), POINTER(c_uint32))


class WjhL1DropAggregateKey(Structure):
    _fields_ = [("ingress_port", c_uint32)]


class WjhL1DropAggregateData(Structure):
    _fields_ = [("is_port_up", c_uint8),
                ("port_down_reason", c_char_p),
                ("description", c_char_p),
                ("state_change_count", c_uint64),
                ("symbol_error_count", c_uint64),
                ("crc_error_count", c_uint64),
                ("timestamp", WjhAggregateTimestamp)]


WjhL1DropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhL1DropAggregateKey), POINTER(WjhL1DropAggregateData), POINTER(c_uint32))

class WjhAggregateFiveTuples(Structure):
    _fields_ = [("sip", c_char_p),
                ("dip", c_char_p),
                ("proto", c_uint8),
                ("sport", c_uint32),
                ("dport", c_uint32)]

class WjhRouterDropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("reason", c_char_p)]

class WjhRouterDropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint32)]


WjhRouterDropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhRouterDropAggregateKey), POINTER(WjhRouterDropAggregateData), POINTER(c_uint32))


class WjhAclDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("binding_point", c_int),
                ("rule_id", c_uint64),
                ("acl_name", c_char_p),
                ("rule", c_char_p),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32),
                ("ingress_lag", c_uint32),
                ("drop_reason", WjhDropReason)]


WjhAclDropRawCb = CFUNCTYPE(c_int, POINTER(WjhAclDropRawInfo), POINTER(c_uint32))


class WjhUserChannelAttr(Structure):
    _fields_ = [("polling_interval", c_uint32),
                ("mode", c_int)]


class RawCallback(Union):
    _fields_ = [("buffer", WjhBufferDropRawCb),
                ("acl", WjhAclDropRawCb),
                ("router", WjhRouterDropRawCb),
                ("L2", WjhL2DropRawCb),
                ("tunnel", WjhTunnelDropRawCb)]


class AggregateCallback(Union):
    _fields_ = [("buffer", c_void_p),
                ("acl", c_void_p),
                ("router", c_void_p),
                ("L2", c_void_p),
                ("L1", WjhL1DropAggregateCb),
                ("tunnel", c_void_p)]


class WjhDropCallbacks(Structure):
    _fields_ = [("drop_reason_group", c_int),
                ("raw_cb", RawCallback),
                ("aggregate_cb", AggregateCallback)]


client_instance = None

class WjhClient(object):

    def __init__(self, conf, logger, platform, is_mlx=True):
        self._conf = conf
        self._logger = logger
        self.wjh_lib = None
        self.portmap = {}
        self.lidmap = {}
        self.agg_cache = {}
        self.sdk_pid = 0
        self.is_mlx = is_mlx
        self.mlx_platform = platform

        signal.signal(signal.SIGTERM, sig_handler)

    def create_user_channel(self, channel_type):
        user_channel = c_uint32()
        if self.wjh_lib.wjh_user_channel_create(channel_type,
                                                byref(user_channel)) != 0:
            self._logger.error("Failed to create user channel")
        else:
            self._logger.info("create channel id %d" % (user_channel.value))

        return user_channel.value


    def get_counter(self, user_channel_id):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_counter_get(user_channel_id,
                    byref(drop_counter), 0) != 0:

            self._logger.error("Failed to get counter")
        else:
            self._logger.info("channel[%d]: drop_counter [%d], recieve_counter [%d]" % (
                              user_channel_id, drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def get_global_counter(self):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_global_counter_get(byref(drop_counter), 0) != 0:
            self._logger.error("Failed to get global counter")
        else:
            self._logger.error("global: drop_counter [%d], recieve_counter [%d]" % (
                               drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def configure_drop_reason_group(self, drop_reason_group, channel_id):
        if self.wjh_lib.wjh_drop_reason_group_bind(drop_reason_group, channel_id) != 0:
            self._logger.error("Failed to bind drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_enable(drop_reason_group, WJH_SEVERITY_ALL_E) != 0:
            self._logger.error("Failed to enable drop reason group")


    def deconfigure_drop_reason_group(self, drop_reason_group):
        if self.wjh_lib.wjh_drop_reason_group_disable(drop_reason_group, WJH_SEVERITY_ALL_E) != 0:
            self._logger.error("Failed to disable drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_unbind(drop_reason_group) != 0:
            self._logger.error("Failed to unbind drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_deinit(drop_reason_group) != 0:
            self._logger.error("Failed to deinit drop reason group")

    def init_latency_thresholds(self):
        # NOTE: this should really be done in CL as switchd init
        #       for now, the config is put under agent as follows
        # netq-agent:
        #     log_level: info
        #     wjh: true
        #     wjh_latency_thresholds:
        #         - tc: 0-4,7
        #           ports: swp5,swp6
        #           hi: 1
        #           lo: 0
        #         - tc: 5
        #           hi: 2
        #           lo: 1

        lat_matrix = []
        for i in range(0, 8):
            lat_matrix.append({'enabled': False, 'tc': i, 'ports': [], 'hi': 0, 'lo': 0})

        tele = '/usr/sbin/netq-wjh-tele'
        cmd = '{} --init'.format(tele)
        try:
            subprocess_wrapper(cmd.split())
        except subprocess.CalledProcessError as e:
            self._logger.error(e)

        lat_thresh = self._conf.get('netq-agent').get('wjh_latency_thresholds', False)
        if lat_thresh:
            #self._logger.info(lat_thresh)
            for lat in lat_thresh:
                tcs = lat.get('tc')
                if not tcs:
                    continue

                ports = []
                if lat.get('ports'):
                    for p in lat.get('ports').split(','):
                        lid = self.lidmap.get(p)
                        if not lid:
                            continue
                        ports.append(lid)
                else:
                    ports = self.portmap.keys()

                tcs = str(tcs) if isinstance(tcs, int) else tcs
                for a_tc in tcs.split(','):
                    r = a_tc.split('-')
                    s = int(r[0])
                    e = s
                    if len(r) > 1:
                        e = int(r[1])
                    if s > 7 or e > 7:
                        continue

                    for i in range(s, e + 1):
                        lat_matrix[i]['enabled'] = True
                        lat_matrix[i]['hi'] = lat.get('hi', 0)
                        lat_matrix[i]['lo'] = lat.get('lo', 0)
                        lat_matrix[i]['ports'] = ports

        for lat in lat_matrix:
            for lid in self.portmap.keys():
                try:
                    clear_cmd = ''
                    if lid not in lat['ports'] or not lat['enabled']:
                        # TBD: ideally we clear the setting that might be in hw
                        #      from the previous incarnation, but sdk will
                        #      generate a lot of error logs if there was no
                        #      prior setting.  The best way is to do all this
                        #      in switchd.
                        # clear_cmd = '--clear'
                        continue
                    cmd = '{} {} --port {} --tc {} --threshold_hi {} --threshold_lo {}'.format(
                            tele, clear_cmd, lid, lat['tc'], lat['hi'], lat['lo'])
                    #self._logger.info(cmd)
                    subprocess_wrapper(cmd.split())
                except subprocess.CalledProcessError as e:
                    self._logger.error(e)


    def latency_monitor_supported(self):
        cmd = '/usr/bin/dpkg -l sx-sdk-eth'
        output = subprocess_wrapper(cmd.split())
        for l in output.splitlines():
            if 'sx-sdk-eth' in l:
                pkg = l.split()[2]
                sdk_ver = pkg.split('.')
                mj = int(sdk_ver[2])
                md = int(sdk_ver[3])
                mi = int(sdk_ver[4].split('-')[0])
                self._logger.info('mlx sdk version: {}.{}.{}'.format(mj, md, mi))
                if (mj == 4 and md == 3 and mi >= 3206) or (mj > 4):
                    return True
        return False

    def buffer_cb_supported(self):
        plat_name = self.mlx_platform.split()[1]
        self._logger.info('PLATFORM: {}'.format(plat_name))
        path = '/usr/share/cumulus-platform/common/platform-info/mellanox.json'
        try:
            with open(path, 'r') as f:
                plat = json.load(f)
                if plat.get(plat_name):
                    plat_json = plat[plat_name][0]
                    asic = plat_json['soc']['model']
                    if asic != 'Spectrum':
                        return True
        except OSError:
            pass

        return False


    def wjh_callback(self, channel_id, aggregation_channel_id):
        global l2_cb, router_cb, tunnel_cb, acl_cb, l1_cb, buffer_cb

        if self.get_port_map() == AGENTSTATUS_FAILURE:
            return

        # L2
        l2_cb = WjhDropCallbacks()
        l2_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L2_E
        l2_cb.raw_cb.L2 = WjhL2DropRawCb(self.wjh_l2_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L2_E, byref(attr), byref(l2_cb)) != 0:
            self._logger.error("Failed to init L2 drop reason group")

        # L3
        router_cb = WjhDropCallbacks()
        router_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ROUTER_E
        router_cb.raw_cb.router = WjhRouterDropRawCb(self.wjh_router_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ROUTER_E, byref(attr), byref(router_cb)) != 0:
            self._logger.error("Failed to init router drop reason group")

        # tunnel
        tunnel_cb = WjhDropCallbacks()
        tunnel_cb.drop_reason_group = WJH_DROP_REASON_GROUP_TUNNEL_E
        tunnel_cb.raw_cb.tunnel = WjhTunnelDropRawCb(self.wjh_tunnel_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_TUNNEL_E, byref(attr), byref(tunnel_cb)) != 0:
            self._logger.error("Failed to init tunnel drop reason group")

        # acl
        acl_cb = WjhDropCallbacks()
        acl_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ACL_E
        acl_cb.raw_cb.acl = WjhAclDropRawCb(self.wjh_acl_drop_raw_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ACL_E, byref(attr), byref(acl_cb)) != 0:
            self._logger.error("Failed to init acl drop reason group")

        # buffer
        if self.buffer_cb_supported():
            buffer_cb = WjhDropCallbacks()
            buffer_cb.drop_reason_group = WJH_DROP_REASON_GROUP_BUFFER_E
            buffer_cb.raw_cb.buffer = WjhBufferDropRawCb(self.wjh_buffer_drop_raw_cb)
            attr = WjhDropReasonGroupAttr()
            attr.attr.buffer_drop.basic_attr.recirculation_port = 0
            attr.attr.buffer_drop.advanced_attr.trap_probability = 8
            if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_BUFFER_E, byref(attr), byref(buffer_cb)) != 0:
                self._logger.error("Failed to init buffer drop reason group")

            if self.latency_monitor_supported():
                self.init_latency_thresholds()

        # L1
        l1_cb = WjhDropCallbacks()
        l1_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L1_E
        l1_cb.aggregate_cb.L1 = WjhL1DropAggregateCb(self.wjh_l1_drop_aggregate_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L1_E, byref(attr), byref(l1_cb)) != 0:
            self._logger.error("Failed to init L1 drop reason group")

        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E, aggregation_channel_id)

        global client_instance
        client_instance = self

        while True:
            time.sleep(15)
            self.flush_agg_cache()


    def wjh_deactive_cb(self):
        self._logger.info("This WJH client is deactivated by another WJH client")
        sys.exit(0)

    def get_port_map(self):
        cmd = 'cat /cumulus/switchd/run/software-tables/root'
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            self._logger.error('Cannot get portmap root')
            return AGENTSTATUS_FAILURE

        table_id = 0
        for l in output.splitlines():
            if 'Backend Port Info (Linux) Table' in l:
                table_id = int(l.split()[1])
                break

        if table_id == 0:
            self.logger.error('Cannot get port map table')
            return AGENTSTATUS_FAILURE

        cmd = 'cat /cumulus/switchd/run/software-tables/{}'.format(table_id)
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            self._logger.error('Cannot get portmap')
            return AGENTSTATUS_FAILURE

        port = ''
        lid = ''
        for l in output.splitlines():
            if 'port-info' in l:
                port = l.split()[1]
            if 'lid' in l:
                lid = l.split()[1]
                self.portmap[lid] = port
                self.lidmap[port] = lid

        return AGENTSTATUS_SUCCESS


    def start_dummy_client(self):
        class _raw_info(object):
            def __init__(self, reason='', drop_type='', ingress_port='',
                         egress_port='', traffic_class=0, acl_rule_id=0,
                         acl_bind_point=0, acl_name='', acl_rule='', smac='',
                         dmac='', sip='', dip='', proto=0, sport=0, dport=0,
                         fts=0.0, lts=0.0, agg_count=0):
                self.drop_agg = WjhDropAggregate()
                self.drop_agg.reason = reason
                self.drop_agg.drop_type = drop_type
                self.drop_agg.ingress_port = ingress_port
                self.drop_agg.egress_port = egress_port
                self.drop_agg.traffic_class = traffic_class
                self.drop_agg.acl_rule_id = acl_rule_id
                self.drop_agg.acl_bind_point = acl_bind_point
                self.drop_agg.acl_name = acl_name
                self.drop_agg.acl_rule = acl_rule
                self.drop_agg.smac = smac
                self.drop_agg.dmac = dmac
                self.drop_agg.sip = sip
                self.drop_agg.dip = dip
                self.drop_agg.proto = proto
                self.drop_agg.sport = sport
                self.drop_agg.dport = dport
                self.drop_agg.agg_count = agg_count
                self.drop_agg.first_timestamp = fts
                self.drop_agg.last_timestamp = lts

        self._logger.info('Started DUMMY wjh agent')
        try:
            data_file = '/etc/netq/.wjh.data'
            with open(data_file, 'r') as f:
                drop_data = yaml.load(f)
                if drop_data:
                    l1_data = drop_data.get('L1-drops')
                    other_data = drop_data.get('Other-drops')

                    t1 = time.time()
                    if other_data:
                        for ddict in other_data:
                            self._logger.debug('--- Others')
                            d = ddict['drop']
                            self._logger.debug('{}'.format(d))
                            fts = t1 - (random.random() * 20)
                            lts = t1 + (random.random() * 20)

                            raw_info = _raw_info(reason=d.get('reason', ''),
                                                 drop_type=d.get('drop_type', ''),
                                                 ingress_port=d.get('ingress_port', ''),
                                                 egress_port=d.get('egress_port', ''),
                                                 traffic_class=0 if not d.get('traffic_class') else int(d.get('traffic_class')),
                                                 acl_rule_id=0 if not d.get('acl_rule_id') else int(d.get('acl_rule_id')),
                                                 acl_bind_point=0 if not d.get('acl_bind_point') else int(d.get('acl_bind_point')),
                                                 acl_name=d.get('acl_name', ''),
                                                 acl_rule=d.get('acl_rule', ''),
                                                 smac=d.get('smac', ''),
                                                 dmac=d.get('dmac', ''),
                                                 proto=0 if not d.get('proto') else int(d.get('proto')),
                                                 sport=0 if not d.get('sport') else int(d.get('sport')),
                                                 dport=0 if not d.get('dport') else int(d.get('dport')),
                                                 sip=d.get('sip', ''),
                                                 dip=d.get('dip', ''),
                                                 fts=fts,
                                                 lts=lts,
                                                 agg_count=0 if not d.get('count') else int(d.get('count')))
                            drop_agg = raw_info.drop_agg
                            key = (drop_agg.reason, drop_agg.ingress_port, drop_agg.sip, drop_agg.dip,
                                   drop_agg.proto, drop_agg.sport, drop_agg.dport, drop_agg.smac,
                                   drop_agg.dmac)
                            self.agg_cache[key] = drop_agg
                            self._logger.debug('--- Others saved')
                    if l1_data:
                        for ddict in l1_data:
                            self._logger.debug('--- L1')
                            d = ddict['drop']
                            self._logger.debug('{}'.format(d))
                            fts = t1 - (random.random() * 20)
                            lts = t1 + (random.random() * 20)

                            L1drop = WjhL1DropAggregate()
                            L1drop.ingress_port = d.get('ingress_port', '')
                            L1drop.is_port_up = d.get('is_port_up') == 'True' or 'False'
                            L1drop.port_down_reason = d.get('port_down_reason', '')
                            L1drop.description = ''
                            if d.get('state_change_count'):
                                L1drop.state_change_count = int(d.get('state_change_count'))
                            if d.get('symbol_error_count'):
                                L1drop.symbol_error_count = int(d.get('symbol_error_count'))
                            if d.get('crc_error_count'):
                                L1drop.crc_error_count = int(d.get('crc_error_count'))
                            L1drop.first_timestamp = fts
                            L1drop.last_timestamp = lts
                            L1drop.save()
                            self._logger.debug('--- L1 saved')

        except Exception as e:
            self._logger.error('Error parsing wjh data file: {}'.format(e))

        while True:
            time.sleep(15)
            self.flush_agg_cache()


    def start_wjh_client(self):
        if not self.is_mlx:
            return self.start_dummy_client()

        try:
            ctypes.CDLL('libsxapi.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxadviser.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxaclhelper.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libscew.so', mode=ctypes.RTLD_GLOBAL)
            self.wjh_lib = ctypes.CDLL('libwjh.so', mode=ctypes.RTLD_GLOBAL)
        except OSError:
            self._logger.error("WJH lib not present")
            return AGENTSTATUS_FAILURE

        #wjh_helper = ctypes.CDLL('libwjhhelper.so', mode=ctypes.RTLD_GLOBAL)

        try:
            cmd = '/usr/bin/pidof sx_sdk'
            sdk_pid = subprocess_wrapper(cmd.split())
            self.sdk_pid = int(sdk_pid)
        except Exception:
            self._logger.debug("sx_sdk not alive")
            return AGENTSTATUS_FAILURE

        global init_params
        init_param = WjhInitParam()
        init_param.force = 1
        init_param.deactive_cb = WjhDeactiveCb(self.wjh_callback)
        init_param.bandwidth_percent = 0
        init_param.log_cb = 0
        init_param.conf_xml_path = 0

        if self.wjh_lib.wjh_init(byref(init_param)) != 0:
            self._logger.error("wjh_init failed")
            return AGENTSTATUS_FAILURE

        channel_id = self.create_user_channel(WJH_USER_CHANNEL_CYCLIC_E)
        aggregation_channel_id = self.create_user_channel(WJH_USER_CHANNEL_AGGREGATE_E)

        self.wjh_callback(channel_id, aggregation_channel_id)

        self._logger.info('WJH DONE, EXITING')
        if self.wjh_lib.wjh_user_channel_destroy(channel_id) != 0:
            self._logger.error("Failed to destroy user channel")

        if self.wjh_lib.wjh_user_channel_destroy(aggregation_channel_id) != 0:
            self._logger.error("Failed to destroy user channel")

        if self.wjh_lib.wjh_deinit() != 0:
            self._logger.error("Failed to deinit wjh service lib")

        self._logger.info("wjh lib is deinitialized")
        return AGENTSTATUS_SUCCESS


    #
    # Raw callback handlers
    #
    def wjh_raw_drop_fill(self, info):
        drop_reason = WjhDropRawInfo()
        drop_reason.reason_id = info.drop_reason.id
        drop_reason.reason = info.drop_reason.reason.decode('utf-8')
        drop_reason.description = info.drop_reason.description
        drop_reason.severity = severity_to_str[info.drop_reason.severity]
        lid = '0x%x' % info.ingress_port
        drop_reason.ingress_port = self.portmap.get(lid) or lid
        drop_reason.timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)

        packet_size = info.packet_size
        packet = list((packet_size * ctypes.c_ubyte).from_address(info.packet))

        drop_reason.packet_size = packet_size
        if (packet_size > 0):
            drop_reason.packet = ''.join('%02x' % i for i in packet)

        return drop_reason


    def parse_7_tuples(self, packet, packet_size):
        tup = {'smac': '00:00:00:00:00:00',
               'dmac': '00:00:00:00:00:00',
               'src': '0.0.0.0',
               'dst': '0.0.0.0',
               'sport': 0,
               'dport': 0,
               'proto': 0}

        if packet_size == 0:
            return tup

        pkt = list((packet_size * ctypes.c_ubyte).from_address(packet))
        try:
            p = raw(binascii.a2b_hex(''.join('%02x' % i for i in pkt)))
            eth = Ether(p)
        except Exception as e:
            return tup

        tup['smac'] = eth.src
        tup['dmac'] = eth.dst

        l3_found = False
        iph = None

        try:
            iph = eth[IP]
            l3_found = True
            tup['proto'] = iph.proto
        except Exception as e:
            self._logger.debug('{}'.format(e))

        if not l3_found:
            try:
                iph = eth[IPv6]
                l3_found = True
                tup['proto'] = iph.nh
            except Exception as e:
                self._logger.debug('{}'.format(e))

        if l3_found:
            tup['src'] = iph.src
            tup['dst'] = iph.dst

            try:
                udph = eth[UDP]
                tup['sport'] = udph.sport
                tup['dport'] = udph.dport
            except Exception as e:
                self._logger.debug('{}'.format(e))

            try:
                tcph = eth[TCP]
                tup['sport'] = tcph.sport
                tup['dport'] = tcph.dport
            except Exception as e:
                self._logger.debug('{}'.format(e))

        self._logger.debug('smac: {}, dmac: {} -- src: {}, dst: {}, proto {}, sport {}, dport {}'.format(
                           tup['smac'], tup['dmac'], tup['src'], tup['dst'], tup['proto'], tup['sport'], tup['dport']))
        return tup


    def wjh_raw_drop_fill_aggregate(self, drop_type, info, egress_port='',
                                    traffic_class=0, acl_rule_id=0,
                                    original_occupancy=0, original_latency=0):
        drop_reason = WjhDropAggregate()
        if drop_type == 'ACL':
            drop_reason.reason = '{}'.format(acl_bp_to_str.get(info.binding_point) or '')
            drop_reason.acl_bind_point = info.binding_point
            drop_reason.acl_name = info.acl_name
            drop_reason.acl_rule = info.rule
        else:
            drop_reason.reason = info.drop_reason.reason.decode('utf-8')
        drop_reason.drop_type = drop_type
        p_tuples = self.parse_7_tuples(info.packet, info.packet_size)
        drop_reason.sip = p_tuples['src']
        drop_reason.dip = p_tuples['dst']
        drop_reason.proto = p_tuples['proto']
        drop_reason.sport = p_tuples['sport']
        drop_reason.dport = p_tuples['dport']
        drop_reason.smac = p_tuples['smac']
        drop_reason.dmac = p_tuples['dmac']
        drop_reason.egress_port = egress_port
        drop_reason.traffic_class = traffic_class
        drop_reason.acl_rule_id = acl_rule_id
        drop_reason.original_occupancy = max(original_occupancy, drop_reason.original_occupancy)
        drop_reason.original_latency = max(original_latency, drop_reason.original_latency)
        lid = '0x%x' % info.ingress_port
        drop_reason.ingress_port = self.portmap.get(lid) or lid

        key = (drop_reason.reason, drop_reason.ingress_port, drop_reason.sip, drop_reason.dip,
               drop_reason.proto, drop_reason.sport, drop_reason.dport, drop_reason.smac,
               drop_reason.dmac)

        agg = self.agg_cache.get(key)
        if agg:
            agg.agg_count += 1
            agg.last_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
        else:
            drop_reason.agg_count = 1
            drop_reason.first_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
            drop_reason.last_timestamp = drop_reason.first_timestamp
            self.agg_cache[key] = drop_reason

        return drop_reason


    def flush_agg_cache(self):
        for key, agg in list(self.agg_cache.items()):
            self._logger.info("Aggregate count: {} first {} last {}".format(agg.agg_count, agg.first_timestamp, agg.last_timestamp))
            agg.save()

        self.agg_cache.clear()


    def wjh_l2_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if raw_info_list_p[i].drop_reason.id in default_filters:
                continue

            self._logger.debug("l2 drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str[raw_info_list_p[i].drop_reason.severity],
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('L2', raw_info_list_p[i])

        return 0


    def wjh_router_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if raw_info_list_p[i].drop_reason.id in default_filters:
                continue
            self._logger.debug("router drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str[raw_info_list_p[i].drop_reason.severity],
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Router', raw_info_list_p[i])

        return 0


    def wjh_tunnel_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("tunnel drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str[raw_info_list_p[i].drop_reason.severity],
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Tunnel', raw_info_list_p[i])

        return 0


    def wjh_buffer_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("buffer drop reason group: ingress port [0x%x], size[%d], reason[%s], occupancy[%d], latency[%d], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].original_occupancy,
                      raw_info_list_p[i].original_latency,
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Buffer', raw_info_list_p[i],
                                   egress_port=raw_info_list_p[i].egress_port,
                                   traffic_class=raw_info_list_p[i].tc,
                                   original_occupancy=raw_info_list_p[i].original_occupancy,
                                   original_latency=raw_info_list_p[i].original_latency)

        return 0


    def wjh_acl_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("acl drop reason group: ingress port [0x%x], size[%d], rule_id[0x%x], binding_point[0x%x], acl_name[%s], rule{%s}, timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                     raw_info_list_p[i].packet_size,
                     raw_info_list_p[i].rule_id,
                     raw_info_list_p[i].binding_point,
                     raw_info_list_p[i].acl_name,
                     raw_info_list_p[i].rule,
                     raw_info_list_p[i].timestamp.tv_sec,
                     raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('ACL', raw_info_list_p[i], acl_rule_id=raw_info_list_p[i].rule_id)
        return 0


    #
    # Aggregate callback handlers
    #
    def wjh_agg_drop_fill(self, key, data):
        drop = WjhDropAggregate()
        drop.reason = key.reason
        drop.sip = key.five_tuples.sip
        drop.dip = key.five_tuples.dip
        drop.proto = key.five_tuples.proto
        drop.sport = key.five_tuples.sport
        drop.dport = key.five_tuples.dport
        drop.agg_count = data.count
        drop.first_timestamp = float(data.timestamp.first_timestamp.tv_sec) + float(
                                     data.timestamp.first_timestamp.tv_nsec / 1000000000.0)
        drop.last_timestamp = float(data.timestamp.last_timestamp.tv_sec) + float(
                                           data.timestamp.last_timestamp.tv_nsec / 1000000000.0)
        return drop


    def wjh_acl_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        # ACL aggregate callback is not very useful as it does not have 5 tuples
        pass


    def wjh_buf_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        # buffer drop aggregate is not useful as it does not have traffic_class
        # info
        pass

    def wjh_l1_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        self._logger.info("")
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            lid = '0x%x' % key_list_p[i].ingress_port
            ingress_port = self.portmap.get(lid) or lid
            '''
            self._logger.info("port %s: is_port_up: %u, port_down_reason: %s, description: %s, state_change_count: %u" %
                  (ingress_port, data_list_p[i].is_port_up, data_list_p[i].port_down_reason,
                   data_list_p[i].description, data_list_p[i].state_change_count))
            self._logger.info("           symbol_error_count: %u, crc_error_count: %u" %
                  (data_list_p[i].symbol_error_count, data_list_p[i].crc_error_count))
            self._logger.info("           timestamp[%ld.%ld, %ld.%ld]" % (data_list_p[i].timestamp.first_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.first_timestamp.tv_nsec, data_list_p[i].timestamp.last_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.last_timestamp.tv_nsec))
            '''
            # L1 drop callback data is quite different from the others
            l1drop = WjhL1DropAggregate()
            l1drop.ingress_port = ingress_port
            l1drop.is_port_up = data_list_p[i].is_port_up
            l1drop.port_down_reason = data_list_p[i].port_down_reason
            l1drop.description = data_list_p[i].description
            l1drop.state_change_count = data_list_p[i].state_change_count
            l1drop.symbol_error_count = data_list_p[i].symbol_error_count
            l1drop.crc_error_count = data_list_p[i].crc_error_count
            l1drop.first_timestamp = float(data_list_p[i].timestamp.first_timestamp.tv_sec) + float(
                                           data_list_p[i].timestamp.first_timestamp.tv_nsec / 1000000000.0)
            l1drop.last_timestamp = float(data_list_p[i].timestamp.last_timestamp.tv_sec) + float(
                                          data_list_p[i].timestamp.last_timestamp.tv_nsec / 1000000000.0)
            olddrops = WjhL1DropAggregate().memmodel()
            olddrop = olddrops.query.get(hostname=HOSTNAME, ingress_port=ingress_port)
            if (olddrop and olddrop.is_port_up == l1drop.is_port_up and
                olddrop.port_down_reason == l1drop.port_down_reason and
                olddrop.state_change_count == l1drop.state_change_count and
                olddrop.symbol_error_count == l1drop.symbol_error_count and
                olddrop.crc_error_count == l1drop.crc_error_count):
                continue

            l1drop.save()

        return 0


    def wjh_rtr_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            '''
            self._logger.info("sip %s, dip %s, proto %u, sport %u, dport %u, reason %s" %
                  (key_list_p[i].five_tuples.sip, key_list_p[i].five_tuples.dip, key_list_p[i].five_tuples.proto, key_list_p[i].five_tuples.sport, key_list_p[i].five_tuples.dport, key_list_p[i].reason))
            self._logger.info("count: %u, timestamp[%ld.%ld, %ld.%ld]" % (data_list_p[i].count,
                                                              data_list_p[i].timestamp.first_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.first_timestamp.tv_nsec, data_list_p[i].timestamp.last_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.last_timestamp.tv_nsec))
            '''
            rtrdrop = self.wjh_agg_drop_fill(key_list_p[i], data_list_p[i])
            rtrdrop.save()
        return 0

    def wjh_l2_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            l2drop = self.wjh_agg_drop_fill(key_list_p[i], data_list_p[i])
            smac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].smac))
            dmac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].dmac))
            l2drop.smac = ''.join('%02x' % i for i in smac)
            l2drop.dmac = ''.join('%02x' % i for i in dmac)
            l2drop.save()

    def wjh_tun_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            tundrop = self.wjh_agg_drop_fill(key_list_p[i], data_list_p[i])
            smac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].smac))
            dmac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].dmac))
            tundrop.smac = ''.join('%02x' % i for i in smac)
            tundrop.dmac = ''.join('%02x' % i for i in dmac)
            tundrop.save()

    def _run():
        pass

    def clean_up(self):
        self._logger.info('WJH clean up')
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E)


def sig_handler(signum, frame):
    if client_instance:
        client_instance.clean_up()
    sys.exit(0)
