# Copyright 2018, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


import subprocess

import time
import eventlet
import eventlet.pools
import greenlet
import sys
import re

from netq_lib.orm.redisdb.models import Dom
from netq_lib.common.enums import Platform
from netq_lib.orm import HOSTNAME

class ReturnGroup:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class DomContext(object):
    DOM_PROCESS_DURATION = 120
    DOM_CMD_TIMEOUT = 20
    LASER_RX_POWER_PATTERN = re.compile(r'(?:Rcvr signal avg optical power|Receiver signal OMA)\((Channel [0-9]+)\)')
    LASER_OUTPUT_POWER_PATTERN = re.compile(r'Transmit avg optical power \((Channel [0-9]+)\)')
    LASER_BIAS_CURRENT_PATTERN = re.compile(r'Laser tx bias current \((Channel [0-9]+)\)')

    def __init__(self):
        self.first_run = True
        self.if_name_list = list()
        self.process_if_list = list()
        self.batch_size = 0
        self.bucket_index = 0
        self.total_buckets = 1
        self.cmd_periodicity = 0
        self.interface_ethtool_op = {}
        self.gthread_dict = {}
        self.sub_process = {}

    def __compute_if_process_list(self, kwargs):
        
        if self.bucket_index == 0:
            self.cmd_periodicity = kwargs.get('period', self.DOM_PROCESS_DURATION)
            if_name_set = kwargs.get('optical_ports', None)
            if not if_name_set:
                # No optical ports to process
                return
            self.if_name_list = list(if_name_set)
            if self.first_run:
                self.total_buckets = 1
                self.batch_size = len(self.if_name_list)
                self.first_run = False
            else:
                self.handle_outdated_objs()
                if self.DOM_PROCESS_DURATION < self.cmd_periodicity:
                    self.total_buckets = (self.cmd_periodicity // self.DOM_PROCESS_DURATION)
                div, mod = divmod(len(self.if_name_list), self.total_buckets)
                self.batch_size = (div+1) if mod > 0 else div

        start_idx = self.bucket_index * self.batch_size
        end_idx = start_idx + self.batch_size
        # Update bucket index
        self.bucket_index += 1
        if self.bucket_index == self.total_buckets:
            self.bucket_index = 0
        # Boundary check for indexes
        if start_idx >= len(self.if_name_list):
            self.process_if_list = list()
            return
        elif end_idx > len(self.if_name_list):
            end_idx = len(self.if_name_list)
        self.process_if_list = self.if_name_list[start_idx:end_idx]

    def process_dom_interfaces(self, kwargs):
        command = "ethtool -m "
        thread_pool = kwargs.get('pool', '*')
        self.__compute_if_process_list(kwargs)

        for ifName in self.process_if_list:
            # Run ethtool command for each interface in a separate thread
            self.__run_command_safe(command + ifName, thread_pool, ifName)

        for ifName in self.process_if_list:
            cmd_time_out = time.time() + int(self.DOM_CMD_TIMEOUT)
            while True:
                cmd_out = ''
                if ifName in self.interface_ethtool_op:
                    cmd_out = self.interface_ethtool_op.pop(ifName)
                    ret_code = 0 if cmd_out.success else -1
                    break

                if time.time() > cmd_time_out:
                    ret_code = -1
                    try:
                        self.sub_process[ifName].terminate()
                        eventlet.sleep(0.3)
                        self.sub_process[ifName].wait()
                        eventlet.kill(gthread_dict[ifName], greenlet.GreenletExit)
                    except Exception as e:
                        break
                    break
                eventlet.sleep(0.3)

            if ret_code == 0:
                eth_m_elems = {}
                for eth_tool_dom in cmd_out.op.splitlines():
                    eth_tool_dom = eth_tool_dom.decode()
                    if ":" not in eth_tool_dom:
                        continue

                    key, value = eth_tool_dom.split(':', 1)
                    key = key.strip()
                    value = value.strip()
                    eth_m_elems[key] = value

                if 'Optical diagnostics support' in eth_m_elems:
                    if 'Yes' not in eth_m_elems['Optical diagnostics support']:
                        continue
                elif 'Transmitter technology' in eth_m_elems:
                    if 'Copper' in eth_m_elems['Transmitter technology']:
                        continue

                dom_obj = self.__create_dom_object(ifName, eth_m_elems)
                dom_obj.save()

    def handle_outdated_objs(self):
        m_dom = Dom.memmodel()
        old_dom_objs = []
        if m_dom:
            old_dom_objs = (
                list(m_dom.query.filter(hostname=HOSTNAME))
            )

        for old_dom_obj in old_dom_objs:
            if old_dom_obj.ifname not in self.if_name_list:
                old_dom_ent = Dom()
                old_dom_ent.copy(old_dom_obj)
                old_dom_ent.delete()

    def __stop_command_checker(self, green_thread, ifName, cmd):
        try:
            (out, err) = green_thread.wait()
            if cmd.returncode != 0:
                self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
            else:
                self.interface_ethtool_op[ifName] = ReturnGroup(success=True, op=out)
            return
        except greenlet.GreenletExit:
            self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
        except Exception:  # pylint: disable=broad-except
            self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
            eventlet.kill(None, *sys.exc_info())

    def __run_command_safe(self, command, thread_pool, ifName):
        """Run time-limited command and update the global proc dict self.interface_ethtool_op
        with the ethtool status and response
        """
        global gthread_dict
        global sub_process
        try:
            cmd = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    
        except (subprocess.CalledProcessError, OSError):
            self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
            return
    
        self.sub_process[ifName] = cmd
        green_thread = thread_pool.spawn(cmd.communicate)
        green_thread.link(self.__stop_command_checker, ifName, cmd)
        self.gthread_dict[ifName] = green_thread

    def __create_dom_object(self, ifname, eth_elems):
        dom_obj = Dom()
        dom_obj.ifname = ifname
        dom_obj.identifier = eth_elems.get('Identifier', 'n/a')
    
        dom_obj.laser_rx_power = {}
        dom_obj.laser_output_power = {}
        dom_obj.laser_bias_current = {}
        if 'Optical diagnostics support' in eth_elems:
            self.__get_single_channel_values(dom_obj, eth_elems)
        elif 'Transmitter technology' in eth_elems:
            self.__get_multi_channel_values(dom_obj, eth_elems)
    
        dom_obj.module_temp = eth_elems.get('Module temperature', 'n/a')
        dom_obj.module_voltage = eth_elems.get('Module voltage', 'n/a')
    
        dom_obj.laser_rx_power_high_alarm_th = eth_elems.get('Laser rx power high alarm threshold', 'n/a')
        dom_obj.laser_rx_power_low_alarm_th = eth_elems.get('Laser rx power low alarm threshold', 'n/a')
        dom_obj.laser_rx_power_high_warning_th = eth_elems.get('Laser rx power high warning threshold', 'n/a')
        dom_obj.laser_rx_power_low_warning_th = eth_elems.get('Laser rx power low warning threshold', 'n/a')
    
        dom_obj.laser_output_power_high_alarm_th = eth_elems.get('Laser output power high alarm threshold', 'n/a')
        dom_obj.laser_output_power_low_alarm_th = eth_elems.get('Laser output power low alarm threshold', 'n/a')
        dom_obj.laser_output_power_high_warning_th = eth_elems.get('Laser output power high warning threshold', 'n/a')
        dom_obj.laser_output_power_low_warning_th = eth_elems.get( 'Laser output power low warning threshold', 'n/a')
    
        dom_obj.laser_bias_current_high_alarm_th = eth_elems.get('Laser bias current high alarm threshold', 'n/a')
        dom_obj.laser_bias_current_low_alarm_th = eth_elems.get('Laser bias current low alarm threshold', 'n/a')
        dom_obj.laser_bias_current_high_warning_th = eth_elems.get('Laser bias current high warning threshold', 'n/a')
        dom_obj.laser_bias_current_low_warning_th = eth_elems.get( 'Laser bias current low warning threshold', 'n/a')
    
        dom_obj.module_temp_high_alarm_th = eth_elems.get('Module temperature high alarm threshold', 'n/a')
        dom_obj.module_temp_low_alarm_th = eth_elems.get('Module temperature low alarm threshold', 'n/a')
        dom_obj.module_temp_high_warning_th = eth_elems.get('Module temperature high warning threshold', 'n/a')
        dom_obj.module_temp_low_warning_th = eth_elems.get('Module temperature low warning threshold', 'n/a')
    
        dom_obj.module_voltage_high_alarm_th = eth_elems.get('Module voltage high alarm threshold', 'n/a')
        dom_obj.module_voltage_low_alarm_th = eth_elems.get('Module voltage low alarm threshold', 'n/a')
        dom_obj.module_voltage_high_warning_th = eth_elems.get('Module voltage high warning threshold', 'n/a')
        dom_obj.module_voltage_low_warning_th = eth_elems.get( 'Module voltage low warning threshold', 'n/a')

        return dom_obj

    @staticmethod
    def __get_single_channel_values(dom_obj, eth_elems):
        dom_obj.laser_rx_power['Channel 1'] = eth_elems.get('Receiver signal average optical power', 'n/a')
        dom_obj.laser_output_power['Channel 1'] = eth_elems.get('Laser output power', 'n/a')
        dom_obj.laser_bias_current['Channel 1'] = eth_elems.get('Laser bias current', 'n/a')
       
    def __get_multi_channel_values(self, dom_obj, eth_elems):
        for key, value in eth_elems.items():
            if 'Channel' in key:
                rx_power = self.LASER_RX_POWER_PATTERN.match(key)
                if rx_power:
                    dom_obj.laser_rx_power[rx_power.group(1)] = value
                else:
                    output_power = self.LASER_OUTPUT_POWER_PATTERN.match(key)
                    if output_power:
                        dom_obj.laser_output_power[output_power.group(1)] = value
                    else:
                        bias_current = self.LASER_BIAS_CURRENT_PATTERN.match(key)
                        if bias_current:
                            dom_obj.laser_bias_current[bias_current.group(1)] = value


def process_dom(output, svc_failed=False, **kwargs):
    platform = kwargs.get('platform', '')
    if not (Platform.VX in platform or Platform.CUMULUS in platform):
        return
    dom.process_dom_interfaces(kwargs)


# Creating Dom object, which holds the context for dom processing
dom = DomContext()

